from __future__ import annotations

from warnings import warn

from dataclasses import dataclass, field
from random import random
from typing import cast, Dict, List, Optional, Tuple, Union

from . import efnlp_pb2 as pb


def split_input(encoded: List[int], num_segments: int, block_size: int) -> List[List[int]]:

    remainder = len(encoded) % num_segments
    normalized_size = int((len(encoded) - remainder) / num_segments)
    if normalized_size <= block_size:
        raise ValueError("too many segments for specified block size")

    segment_sizes = [normalized_size + (1 if i < remainder else 0) for i in range(num_segments)]

    s, e = 0, segment_sizes[0]
    segments = []
    for i in range(num_segments):
        segments.append(encoded[max(0, s - block_size) : e])
        s += segment_sizes[i]
        if i == num_segments - 1:
            e = len(encoded)
        else:
            e += segment_sizes[i + 1]

    return segments


def merge_split_input(segments: List[List[int]], block_size: int) -> List[int]:
    recomposed = segments[0]
    for i in range(1, len(segments)):
        recomposed += segments[i][block_size:]
    return recomposed


@dataclass
class Language:

    size: int
    stot: Dict[str, int]
    ttos: Dict[int, str]

    @staticmethod
    def from_proto_file(name: str) -> Language:
        with open(name, "rb") as f:
            return Language.from_proto_bytes(f.read())

    @staticmethod
    def from_proto_bytes(data: bytes) -> Language:
        P = pb.Language()
        P.ParseFromString(data)
        return Language.from_proto(P)

    @staticmethod
    def from_proto(L: pb.Language) -> Language:
        return Language(
            size=len(L.lang),
            stot={e.data: e.token for e in L.lang},
            ttos={e.token: e.data for e in L.lang},
        )

    def proto(self) -> pb.Language:
        L = pb.Language()
        for t, s in self.ttos.items():
            e = pb.Encoder(token=t, data=s)
            L.lang.append(e)
        return L

    def encode(self, s: str) -> List[int]:
        raise NotImplementedError("Base Language doesn't encode (yet)")

        # General: incrementally parse forward with the longest match
        # in a prefix tree?
        #
        # i = 0
        # while i is not None:
        #   t, i = self.encoder.encode(s, i) # token and next char (i + len(match) | None)
        #
        # what's an "encoder"? 
        # 
        # class Encoder:
        #   depth: int
        #   prefixes: Dict[str, Encoder]
        #   
        #   def encode(self, s: str, i: int) -> Tuple[int, Optional[int]]:
        #       if i >= len(s):
        #           raise ValueError("cannot encode past string length")
        #       if s[i] not in self.prefixes:
        #           return None
        #       self.prefixes.encode(s, i)
        #           
        # 

    def decode(self, tokens: Union[int, List[int]]) -> str:
        if isinstance(tokens, list):
            return "".join(self.ttos[t] for t in tokens)
        return self.decode([tokens])


@dataclass
class CharLanguage(Language):
    @staticmethod
    def from_corpus(C: str) -> CharLanguage:
        lang = sorted(list(set(c for c in C)))
        return CharLanguage(
            size=len(lang),
            stot={c: i for i, c in enumerate(lang)},
            ttos={i: c for i, c in enumerate(lang)},
        )

    @staticmethod
    def from_proto_file(name: str) -> CharLanguage:
        with open(name, "rb") as f:
            return CharLanguage.from_proto_bytes(f.read())

    @staticmethod
    def from_proto_bytes(data: bytes) -> CharLanguage:
        P = pb.Language()
        P.ParseFromString(data)
        return CharLanguage.from_proto(P)

    @staticmethod
    def from_proto(L: pb.Language) -> CharLanguage:
        return CharLanguage(
            size=len(L.lang),
            stot={e.data: e.token for e in L.lang},
            ttos={e.token: e.data for e in L.lang},
        )

    def encode(self, s: str) -> List[int]:
        return [self.stot[c] for c in s]


class Sampler:

    # total: int
    # counts: Dict[int, int]

    def __init__(self, total: Optional[int] = 0, counts: Optional[Dict[int, int]] = None) -> None:
        self.total: int = total or 0
        self.counts: Dict[int, int] = {} if counts is None else {t: c for t, c in counts.items()}

    @staticmethod
    def from_proto_file(name: str) -> Sampler:
        with open(name, "rb") as f:
            return Sampler.from_proto_bytes(f.read())

    @staticmethod
    def from_proto_bytes(data: bytes) -> Sampler:
        assert isinstance(data, bytes)
        P = pb.Sampler()
        P.ParseFromString(data)
        return Sampler.from_proto(P)

    @staticmethod
    def from_proto(S: pb.Sampler) -> Sampler:
        assert isinstance(S, pb.Sampler), (type(S), S)
        s = Sampler(total=S.total)
        s.counts = {stc.token: stc.count for stc in S.counts}
        return s

    @staticmethod
    def copy(S: Sampler) -> Sampler:
        return Sampler(total=S.total, counts=S.counts)

    def proto(self) -> pb.Sampler:
        S = pb.Sampler(total=self.total)
        for t, c in self.counts.items():
            stc = pb.SamplerTokenCount(token=t, count=c)
            S.counts.append(stc)
        return S

    def __bytes__(self) -> bytes:
        return self.proto().SerializeToString()

    def __contains__(self, token: int) -> bool:
        return token in self.counts

    def __add__(self, t: int) -> Sampler:
        return self.add(t)

    def __iadd__(self, t: int) -> Sampler:
        return self.add(t)

    def __getitem__(self, t: int) -> int:
        return self.counts[t]

    def get(self, t: int, default: int = 0) -> int:
        return self.counts.get(t, default)

    def items(self):  # -> List[Tuple[int, int]]:
        return self.counts.items()

    def __iter__(self):
        return self.counts.__iter__()

    def __next__(self):
        return self.counts.__next__()

    def add(self, t: int) -> Sampler:
        self.total += 1
        if t in self.counts:
            self.counts[t] += 1
        else:
            self.counts[t] = 1
        return self

    def memory(self) -> int:
        return 4 + 2 * len(self.counts)

    def merge(self, other: Sampler) -> Sampler:
        self.total += other.total
        for t in self.counts:
            self.counts[t] += other.get(t)
        for t in other:
            if t not in self.counts:
                self.counts[t] = other[t]
        return self

    def sample(self) -> int:
        r = self.total * random()
        for t, c in self.counts.items():
            if r < c:
                return t
            r -= c
        raise Exception("We should not be here, sampler malformed")

    def probability(self, token: int) -> float:
        return self.counts.get(token, 0) / self.total


@dataclass
class SuffixTree:

    token: int  # token index
    depth: int = field(init=False)  # occurrences
    sampler: Sampler = field(init=False)
    children: Dict[int, SuffixTree] = field(init=False)  # "children" is a bad name

    def __post_init__(self) -> None:
        self.depth = 0
        self.sampler = Sampler()
        self.children = {}

    @staticmethod
    def from_proto_file(name: str) -> SuffixTree:
        with open(name, "rb") as f:
            return SuffixTree.from_proto_bytes(f.read())

    @staticmethod
    def from_proto_bytes(data: bytes) -> SuffixTree:
        P = pb.SuffixTree()
        P.ParseFromString(data)
        return SuffixTree.from_proto(P)

    @staticmethod
    def from_proto(T: pb.SuffixTree) -> SuffixTree:
        t = SuffixTree(token=T.token)
        t.sampler = Sampler.from_proto(T.sampler)
        for c in T.prefixes:
            t.children[c.token] = SuffixTree.from_proto(c)
        return t

    @staticmethod
    def copy(T: SuffixTree) -> SuffixTree:
        S = SuffixTree(token=T.token)
        S.depth = T.depth
        S.sampler = Sampler.copy(T.sampler)
        S.children = {t: SuffixTree.copy(T.children[t]) for t in T}
        return S

    def proto(self) -> pb.SuffixTree:
        T = pb.SuffixTree(token=self.token)
        T.sampler.CopyFrom(self.sampler.proto())
        for _, c in self.children.items():
            T.prefixes.append(c.proto())
        return T

    def __bytes__(self) -> bytes:
        return self.proto().SerializeToString()

    def __add__(self, child: SuffixTree) -> SuffixTree:
        return self.add(child)

    def __contains__(self, token: int) -> bool:
        return token in self.children

    def __getitem__(self, token: int) -> SuffixTree:
        return self.children[token]

    def __setitem__(self, token: int, child: SuffixTree) -> SuffixTree:
        self.children[token] = child
        return self

    def __iter__(self):
        return self.children.__iter__()

    def __next__(self):
        return self.children.__next__()

    def __call__(self, prefix: List[int]) -> float:
        """Convenience for sampling."""
        return self.sample(prefix)

    def get(self, token: int) -> Optional[SuffixTree]:
        """Return the child with given token, or None."""
        return self.children.get(token)

    def add(self, child: SuffixTree) -> SuffixTree:
        """Add a "child" token in the children map (though "children"
        is a bit misleading in context). Chainable."""
        self.children[child.token] = child
        return self

    def parse(self, prefix: List[int], successor: int) -> SuffixTree:
        """Parse a prefix into this SuffixTree, including the successor
        token. Chainable."""
        self.sampler += successor
        if len(prefix) > 0:
            C = self.get(prefix[-1])
            if C is None:
                C = SuffixTree(prefix[-1])
                self[prefix[-1]] = C
            C.parse(prefix[:-1], successor)
            self.depth = max(self.depth, C.depth + 1)

        return self

    @staticmethod
    def merge_proto_bytes(a: bytes, b: bytes) -> SuffixTree:
        return SuffixTree.merge_trees(
            SuffixTree.from_proto_bytes(a), SuffixTree.from_proto_bytes(b)
        )

    @staticmethod
    def merge_proto(a: pb.SuffixTree, b: pb.SuffixTree) -> SuffixTree:
        return SuffixTree.merge_trees(SuffixTree.from_proto(a), SuffixTree.from_proto(b))

    @staticmethod
    def merge_trees(a: SuffixTree, b: SuffixTree) -> SuffixTree:
        return a.merge(b)

    def merge(self, other: SuffixTree) -> SuffixTree:
        """Merge another SuffixTree into this one"""
        assert self.token == other.token
        self.depth += max(self.depth, other.depth)
        self.sampler.merge(other.sampler)
        for t in self:
            if t in other:
                self.children[t].merge(other.children[t])
        for t in other:
            if t not in self:
                self.children[t] = SuffixTree.copy(other.children[t])
                # TODO: is a copy required? why not just reference?
        return self

    def memory(self) -> int:
        """Return an estimate of the memory requirements for this tree"""
        return self.sampler.memory() + sum([4 + c.memory() for t, c in self.children.items()])

    def prefixes(self) -> List[List[int]]:
        """List all prefixes held in this SuffixTree. These are
        basically the set of paths to leaf nodes from this node."""
        if len(self.children) == 0:
            return [[self.token]]
        return [(p + [self.token]) for _, c in self.children.items() for p in c.prefixes()]

    def patterns(self) -> List[Tuple[List[int], int]]:
        """List all "patterns" held in this SuffixTree. These are
        basically the set of paths to leaf nodes from this node,
        along with any successor tokens."""
        if len(self.children) == 0:
            return [([self.token], t) for t in self.sampler.counts]
        return [
            (p[0] + [self.token], p[1]) for _, c in self.children.items() for p in c.patterns()
        ]

    def search(self, prefix: List[int]) -> List[int]:
        """Search for a given prefix in this tree. Returns
        the longest match."""
        if (len(prefix) > 0) and (prefix[-1] in self):
            if len(prefix) == 1:
                return [prefix[-1], self.token]
            return self[prefix[-1]].search(prefix[:-1]) + [self.token]
        return [self.token]

    def match(self, prefix: List[int]) -> bool:
        """Return true if the presented prefix is in the suffix tree"""
        if len(prefix) == 0:
            raise ValueError("can't match the empty token string")
        if prefix[-1] in self:
            if len(prefix) == 1:
                return True
            return self[prefix[-1]].match(prefix[:-1])
        return False

    def sample(self, prefix: List[int]) -> int:
        """Sample a token that likely follows a prefix. Requires
        a normalized tree for now. Recurses to search for the
        longest matching suffix for the prefix."""
        if len(prefix) == 0 or prefix[-1] not in self:
            return self.sampler.sample()
        return self[prefix[-1]].sample(prefix[:-1])

    def probability(self, prefix: List[int], successor: int) -> float:
        """Return pattern probability. Basically suffix search
        but returning the probability of the longest match."""

        if len(prefix) == 0:
            return self.sampler.counts.get(successor, 0) / self.sampler.total

        # search children
        C = self.children.get(prefix[-1])
        if C is None:
            return self.sampler.counts.get(successor, 0) / self.sampler.total

        # recurse into matched child with the prefix's prefix
        return C.probability(prefix[:-1], successor)

        # TODO: verify


class SuffixTreeSet:

    # size: int  # "vocab" size
    # depth: int = field(init=False)
    # trees: Dict[int, SuffixTree] = field(init=False)
    # sampler: Sampler = field(init=False) # for counting raw token occurrences

    def __init__(self, size: int = 0) -> None:
        self.size: int = size
        self.depth: int = 0
        self.trees: Dict[int, SuffixTree] = {t: SuffixTree(t) for t in range(self.size)}
        self.sampler: Sampler = Sampler()

    @staticmethod
    def from_proto_file(name: str) -> SuffixTreeSet:
        with open(name, "rb") as f:
            return SuffixTreeSet.from_proto_bytes(f.read())

    @staticmethod
    def from_proto_bytes(data: bytes) -> SuffixTreeSet:
        P = pb.SuffixTreeSet()
        P.ParseFromString(data)
        return SuffixTreeSet.from_proto(P)

    @staticmethod
    def from_proto(S: pb.SuffixTreeSet) -> SuffixTreeSet:
        s = SuffixTreeSet(size=len(S.prefixes))
        for c in S.prefixes:
            s.trees[c.token] = SuffixTree.from_proto(c)
        return s

    def proto(self) -> pb.SuffixTreeSet:
        S = pb.SuffixTreeSet()
        for _, c in self.trees.items():
            S.prefixes.append(c.proto())
        return S

    def __bytes__(self) -> bytes:
        return self.proto().SerializeToString()

    def __contains__(self, token: int) -> bool:
        return token in self.trees

    def __getitem__(self, token: int) -> SuffixTree:
        return self.trees[token]

    def __iter__(self):
        return self.trees.__iter__()

    def __next__(self):
        return self.trees.__next__()

    def items(self):
        return self.trees.items()

    def parse(self, prefix: List[int], successor: int) -> SuffixTreeSet:
        if len(prefix) == 0:
            raise ValueError("Cannot parse empty prefix")
        self.sampler += successor
        if prefix[-1] not in self.trees:
            self.trees[prefix[-1]] = SuffixTree(prefix[-1])
        self.trees[prefix[-1]].parse(prefix[:-1], successor)
        self.depth = max(self.depth, self.trees[prefix[-1]].depth + 1)
        return self

    def parse_all(
        self, tokens: List[int], block_size: int, ignore_start: bool = False
    ) -> SuffixTreeSet:
        if not ignore_start:
            for i in range(1, block_size):  # leading sub-block_size prefixes
                self.parse(tokens[0:i], tokens[i])
        for i in range(block_size, len(tokens) - 1):
            self.parse(tokens[i - block_size : i], tokens[i])
        return self

    def merge(self, other: SuffixTreeSet) -> SuffixTreeSet:
        assert self.size == other.size
        self.depth = max(self.depth, other.depth)
        for i in range(self.size):
            self.trees[i].merge(other.trees[i])
        return self

    def memory(self) -> int:
        return self.sampler.memory() + sum([T.memory() for _, T in self.trees.items()])

    def prefixes(self, token: Optional[int] = None) -> List[List[int]]:
        if token is None:
            r = []
            for t in range(self.size):
                r += self.prefixes(t)
            return r
        return self.trees[token].prefixes()

    def patterns(self, token: Optional[int] = None) -> List[Tuple[List[int], int]]:
        if token is None:
            r = []
            for t in range(self.size):
                r += self.patterns(t)
            return r
        return self.trees[token].patterns()

    def search(self, prefix: List[int]) -> List[int]:
        if len(prefix) == 0:
            raise ValueError("can't search the empty token string")
        if prefix[-1] in self:
            return self.trees[prefix[-1]].search(prefix[:-1])
        return []

    def match(self, prefix: List[int]) -> bool:
        if len(prefix) == 0:
            raise ValueError("can't match the empty token string")
        if prefix[-1] in self:
            return self.trees[prefix[-1]].match(prefix[:-1])
        return False

    def sample(self, prefix: List[int] = []) -> int:
        if (len(prefix) == 0) or (prefix[-1] not in self.trees):
            return self.sampler.sample()  # sample from raw token occurrence frequencies
        return self.trees[prefix[-1]].sample(prefix[:-1])

    def probability(self, prefix: List[int], successor: int) -> float:
        if (len(prefix) == 0) or (prefix[-1] not in self.trees):
            return self.sampler.probability(
                successor
            )  # raw occurrence probability of the successor
        return self.trees[prefix[-1]].probability(prefix[:-1], successor)

    def generate(self, size: int, window: int, prompt: List[int]) -> List[int]:
        code = prompt
        for i in range(size):
            prefix = code if len(code) < window else code[-window:]
            code.append(self.sample(prefix))
        return code


AccmType = Tuple[int, bytes]


class _SuffixTreeParser:  # (beam.DoFn):
    def __init__(
        self,
        language: CharLanguage,  # is beam so specific about this?
        block_size: int,
        documents: bool = False,
        doc_start_tok: int = 0,
        doc_end_tok: int = 1,
        text_encoded: bool = False,
        text_delimiter: str = ",",
        text_field: str = "segment",
    ):
        self.L = language
        self.B = block_size

        # TODO: annoying; setattr over generic kwargs? cleaner code
        # but much more opaque interface/API here
        self.documents = documents
        self.doc_start_tok = doc_start_tok
        self.doc_end_tok = doc_end_tok
        self.text_encoded = text_encoded
        self.text_delimiter = text_delimiter
        self.text_field = text_field

    def process(self, record):  # -> Generator[Tuple[int, bytes]]?

        # size: int; but what about dynamic parsing?
        T = SuffixTreeSet()

        # encode segment in language; TODO: generalize
        encoded: List[int] = []
        text = cast(str, record[self.text_field])
        if self.text_encoded:
            encoded = [int(i) for i in text.split(self.text_delimiter)]
        encoded = self.L.encode(text)

        # TODO: doc start + encoded + doc end...
        if self.documents:
            encoded = [self.doc_start_tok] + encoded + [self.doc_end_tok]

        # parse token string
        T.parse_all(encoded, self.B)

        # emit each tree
        for t, tree in T.trees.items():
            # TODO: tuple form sufficient for a "keyed" PCollection?
            yield (t, bytes(tree))  # bytes -> protobuf encoding


class _SuffixTreeMerge:  # (beam.CombineFn):
    # are we merging based on proto, or pyobjects?

    def create_accumulator(self) -> Optional[AccmType]:
        return None  # SuffixTree() but for what token?

    def add_input(self, accm: Optional[AccmType], inp: AccmType) -> AccmType:
        if accm is None:
            return inp
        tree = SuffixTree.merge_proto_bytes(accm[1], inp[1])
        return (tree.token, bytes(tree))

    # TODO: List/Iterable arg? or varargs?
    def merge_accumulators(self, accumulators: List[Optional[AccmType]]) -> Optional[AccmType]:
        tree = None
        for accm in accumulators:
            if accm:
                assert isinstance(accm, tuple), print(type(accm), accm)
                t, b = accm  # split into token/bytes
                if tree:
                    other = SuffixTree.from_proto_bytes(b)
                    tree.merge(other)  # won't merge if token mismatched
                else:
                    tree = SuffixTree.from_proto_bytes(b)
        if tree:
            return (tree.token, bytes(tree))
        return None

    def extract_output(self, accm: AccmType) -> AccmType:
        return accm


try:

    import apache_beam as beam

    class SuffixTreeParser(_SuffixTreeParser, beam.DoFn):  # MRO requires _SuffixTreeParser first
        pass

    class SuffixTreeMerge(
        _SuffixTreeMerge, beam.CombineFn
    ):  # MRO requires _SuffixTreeParser first
        pass

except ImportError as err:

    warn(
        f"Cannot import beam: SuffixTreeParser and SuffixTreeMerge will not be parallelized ({err})"
    )

    class Unimportable:
        def __init__(self, *args, **kwargs) -> None:
            raise NotImplementedError(
                f"Cannot initialize {self.__class__.__name__} without `beam` modules"
            )

    class SuffixTreeParser(_SuffixTreeParser):  # type:ignore
        pass

    class SuffixTreeMerge(_SuffixTreeMerge):  # type:ignore
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['efnlp']

package_data = \
{'': ['*']}

install_requires = \
['protobuf>=3.19,<4.0']

setup_kwargs = {
    'name': 'efnlp',
    'version': '0.1.9',
    'description': 'Empirical Frequency Natural Language Processing',
    'long_description': '# efnlp\n\nThis is the _jankiest possible_ take at computing n-gram or conditional empirical frequency ((C)EF) distributions for successor tokens given a fixed-width prefix. Yes that\'s just a Markov model. In a way this is _purposefully_ naive, and interested in how far such a naive technique goes. \n\nThere\'s `python` and `c++` code (in [`efnlp`](/efnlp) and [`cpp`](/cpp) respectively) for analyzing text and creating sampling datastructures. There\'s also `go` code (in [`go`](/go)) for a gRPC server to read sampling datastructures and serve batch or streaming generated text. \n\n## Motivation (even though we just playin)\n\nThe basic idea is to explore computing these values and using them in generative sampling tasks. We "hypothesize" that the asymptotic (in data and model class universality) result of computational [(L)LMs](https://en.wikipedia.org/wiki/Language_model) is a high-fidelity representation of the (C)EF in the training data. This is a bit tautologous, for a consistent statistical model estimated on sufficient data. See `paper/*.pdf` for more; however I have had repeated personal experiences with really cool complicated models basically reducing to the information in counts and averages, much like (C)EFs. \n\nOf course, a (C)EF "model" should have serious generalizability challenges when confronted with valid sequences not appearing in the training data. Particularly if we wish to generate text (via token sequences), we may generate sequences purely by chance that have no (or limited) precedent and thus have limited predictive capacity. (A Markov model can always reduce to token occurrence frequencies, though randomly choosing tokens is far from language modeling.) This is a primary goal of statistical models: to find (or use) structure to "fill in the gaps" of a particular corpus of training data. However, as a training corpus grows to encompass more and more of the possibilities of discourse, the likelihood we have not "seen" a valid sequence should also decrease, and a model should also be getting progressively more aligned with the (C)EFs. \n\nAlso we would expect the required storage for a purely (C)EF "model" to grow quite large. This is, in principal, another primary goal of statistical models: to "compress" statistical information without diminishing validation/generalization performance. Yet, the SOTA statistical models are themselves quite extraordinarily large; samples are not reasonable computable on a single machine, involving many billions of parameters and linear algebra involved enough to require specialized hardware. \n\nObviously if simple (C)EF models were of comparable quality to modern LLMs they would be making all the noise instead. This isn\'t about a better approach, just a personal investigation. \n\n## Example \n\nYou can run the `python` code from the CLI as\n```shell\n$ python -m efnlp -c data/tinywillspeare.txt -m -s -b 10 -g 100000 -o sample-results.txt\n[2023-01-28T21:37:56.223686 | 0.000020s] Reading text corpus\n[2023-01-28T21:37:56.225040 | 0.001374s] Forming (character) language\n[2023-01-28T21:37:56.273636 | 0.049969s] Encoding corpus in the language constructed\n[2023-01-28T21:37:56.345648 | 0.122002s] Corpus is 1,115,393 tokens long\n[2023-01-28T21:37:56.345711 | 0.122036s] Parsing prefix/successor tokens\n[2023-01-28T21:38:24.269634 | 28.045963s] Parsed prefixes and successors in corpus in 27923.90ms\n[2023-01-28T21:38:27.680794 | 31.457128s] Found 858,920 prefixes (77.0% of possible)\n[2023-01-28T21:38:34.836321 | 38.612656s] Found 937,254 patterns (84.0% of possible)\n[2023-01-28T21:38:37.554198 | 41.330531s] Memory (roughly) required: 31.9 MB (about 4,177,347 dbl, 8,354,695 fl)\n[2023-01-28T21:38:37.554237 | 41.330562s] Sampling and decoding 100000 tokens\n[2023-01-28T21:38:38.493836 | 42.270165s] Generation frequency: 9.4 us/tok\n[2023-01-28T21:38:38.493869 | 42.270194s] Writing sampled results to sample-results.txt\n```\n(for which you can see generated text in [`results`](/sample-results.txt)). The compiled `c++` is similar, \n```shell\ncpp$ ./efnlp++ -c data/tinywillspeare.txt -m -s -b 10 -g 100000 -o sample-results.txt\n```\nor the `go`\n```shell\ngo$ go run *.go -parse \\\n\t-input ../data/tinywillspeare.txt \\\n\t-language ../cpp/language.proto.bin \\\n\t-block 10 \\\n\t-generate 10000 \\\n\t-print=false\n```\n(note the use of a `language` datastructure in `proto` format). \n\nOur "model" here (in `python`) is, more or less, 8M `double`s worth of "parameters" and "trains" (estimates) in a single process on an (old) macbook in under a minute (for 10-token sequence statistics). Sampling is basically constant time, relying on hashmaps; the example above takes about 0.1ms per character sampled (with zero code optimizations). The (C)EF "model" are a significant inflation of the data size: 1.1MB of data turns into 62.4MB of statistics. But honestly the results aren\'t that bad. It\'s junk of course, but on the surface comparable to generative text from a 10M parameter transformer style model applied to the same dataset that trained for 15 minutes on a GPU ([tweet here](https://twitter.com/karpathy/status/1615400286293753856?cxt=HHwWgIDUqY2Ah-ssAAAA), [code here](https://github.com/karpathy/nanoGPT)). \n\nThe full CLI arg list can be reviewed from\n```shell\n$ python -m efnlp -h\n```\n\nMore comprehensive results are detailed in the following tables (see `paper/*.pdf` for more discussion): \n\n| B | \\# prefixes | unique pref. | \\# pattern | unique patt. | avg succ./pref. | memory |\n| --- | --- | --- | --- | --- | --- | --- |\n|  1 | 65 | 0.0\\% | 1,403 | 0.1\\% | 21.6 | 3kB |\n|  2 | 1,403 | 0.1\\% | 11,556 | 1.0\\% | 8.2 | 36kB |\n|  3 | 11,556 | 1.0\\% | 50,712 | 4.5\\% | 4.4 | 221kB |\n|  4 | 50,712 | 4.5\\% | 141,021 | 12.6\\% | 2.8 | 876kB |\n|  5 | 141,021 | 12.6\\% | 283,313 | 25.4\\% | 2.0 | 2.5MB |\n|  7 | 447,352 | 40.1\\% | 609,659 | 54.7\\% | 1.4 | 10.1MB |\n| 10 | 858,920 | 77.0\\% | 937,254 | 84.0\\% | 1.1 | 31.9MB |\n| 12 | 991,391 | 88.9\\% | 1,027,857 | 92.2\\% | 1.0 | 50.4MB |\n| 15 | 1,069,423 | 95.9\\% | 1,081,060 | 96.9\\% | 1.0 | 80.6MB |\n| 20 | 1,103,358 | 98.9\\% | 1,106,345 | 99.2\\% | 1.0 | 133MB |\n\nWhat follows are comparative parsing/generation speeds (for sampling 1M tokens):\n\n|  B  | py parse | py t/&tau; | go parse | go t/&tau; | c++ parse | c++ t/&tau; |\n| --- | ------ | --------- | ----- | -------- | ----- | -------- |\n|  1  | 1.0s   |  1.4&mu;s |  49ms | 0.1&mu;s | 131ms | 0.1&mu;s |\n|  2  | 2.0s   |  1.7&mu;s | 109ms | 0.1&mu;s | 248ms | 0.2&mu;s |\n|  3  | 3.3s   |  2.1&mu;s | 222ms | 0.2&mu;s | 419ms | 0.3&mu;s |\n|  4  | 4.3s   |  2.6&mu;s | 361ms | 0.3&mu;s | 612ms | 0.4&mu;s |\n|  5  | 6.4s   |  3.2&mu;s | 585ms | 0.5&mu;s |  1.1s | 0.5&mu;s |\n|  7  | 12.0s  |  4.9&mu;s |  1.2s | 0.7&mu;s |  2.0s | 0.7&mu;s |\n| 10  | 28.0s  |  7.0&mu;s |  2.6s | 0.9&mu;s |  1.9s | 0.8&mu;s |\n| 12  | 37.3s  |  8.3&mu;s |  4.1s | 1.1&mu;s |  2.5s | 1.0&mu;s |\n| 15  | 54.3s  |  9.7&mu;s |  5.2s | 1.2&mu;s |  3.2s | 1.0&mu;s |\n| 20  | 129.0s | 12.7&mu;s |  8.4s | 1.6&mu;s |  4.4s | 1.3&mu;s |\n\n\nNote (until we get better formatting) that the repeated "parse" (parse time) and "t/&tau;" (time per token) columns are for `python`, `go`, and `c++` respectively. With each we can generate text at O(&mu;s)/token. However compiled codes (`go` and `c++`) are faster (roughly by an order of magnitude) and seem to scale a bit better with longer sequences. With compiled code we can parse out the (C)EFs in shakespeare in seconds; while `python` is still "reasonable" it takes a couple minutes to parse 20-token sequences. \n\n## Usage\n\n### python\n\nSee [`efnlp/__main__.py`](/efnlp/__main__.py) for an example of running the no-doubt-dumb implementations in [`efnlp/__init__.py`](/efnlp/__init__.py). \n\n### c++\n\nSee [`cpp/*`](/cpp), with stuff that should be ready for `cmake` to run through a build. Embarassingly, pretty much everything is just in [`main.cpp`](/cpp/main.cpp). \n\n### go\n\nSee [`go/*`](/go), with stuff that should be ready to build. There\'s a very brief [`README.md`](/go/README.md) \n\n',
    'author': 'W. Ross Morrow',
    'author_email': 'morrowwr@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
